from rest_framework.serializers import BooleanField
from rest_framework.request import Request
from .permissions import validate_hard_delete_is_allowed


def is_hard_delete_request(request: Request) -> bool:
    permanent = request.query_params.get('permanent')
    is_hard_delete = BooleanField(required=False, allow_null=True).run_validation(permanent)

    if is_hard_delete:
        validate_hard_delete_is_allowed()

    return is_hard_delete



from django.db.models import QuerySet
from django.utils import timezone


class SoftDeleteQuerySet(QuerySet):
    def only_deleted(self):
        return self.filter(deleted_at__isnull=False)

    def without_deleted(self):
        return self.filter(deleted_at__isnull=True)

    #  bulk deleting
    def delete(self, hard: bool = False):
        if hard:
            return super(SoftDeleteQuerySet, self).delete()

        return super(SoftDeleteQuerySet, self).update(deleted_at=timezone.now())

    #  bulk restore
    def restore(self):
        return super(SoftDeleteQuerySet, self).update(deleted_at=None)



from django.db import models
from django.utils import timezone
from .managers import SoftDeleteManager


class Timestampable(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class SoftDeletes(models.Model):
    deleted_at = models.DateTimeField(null=True)

    objects = SoftDeleteManager()
    objects_deleted = SoftDeleteManager(only_deleted=True)
    objects_with_deleted = SoftDeleteManager(with_deleted=True)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False, hard: bool = False):
        if hard:
            return super().delete(using, keep_parents)

        self.deleted_at = timezone.now()
        return self.save()

    def restore(self):
        self.deleted_at = None
        return self.save()


class Model(Timestampable, SoftDeletes, models.Model):
    class Meta:
        abstract = True



from rest_framework import viewsets

from .mixins import (
    ListDeletedModelMixin,
    ListWithDeletedModelMixin,
    RetrieveDeletedModelMixin,
    RestoreModelMixin,
    BulkRestoreModelMixin,
    DestroyModelMixin,
    BulkDestroyModelMixin
)


class SoftDeleteModelViewSet(ListDeletedModelMixin,
                             ListWithDeletedModelMixin,
                             RetrieveDeletedModelMixin,
                             RestoreModelMixin,
                             BulkRestoreModelMixin,
                             DestroyModelMixin,
                             BulkDestroyModelMixin,
                             viewsets.ModelViewSet):
    pass



from django.conf import settings
from rest_framework.exceptions import PermissionDenied


def validate_hard_delete_is_allowed():
    if not getattr(settings, 'ALLOW_PERMANENT_BULK_DELETE', False):
        raise PermissionDenied()



from rest_framework import routers


class DefaultRouter(routers.DefaultRouter):
    def __init__(self, *args, **kwargs):
        super(DefaultRouter, self).__init__(*args, **kwargs)

    routers.DefaultRouter.routes[0].mapping.setdefault('delete', 'bulk_destroy')

    # list deleted
    routers.DefaultRouter.routes.insert(0, routers.Route(
        url=r'^{prefix}/deleted{trailing_slash}$',
        mapping={
            'get': 'list_deleted',
        },
        name='{basename}-deleted',
        detail=False,
        initkwargs={'suffix': 'Deleted'}
    ))

    # list with deleted
    routers.DefaultRouter.routes.insert(1, routers.Route(
        url=r'^{prefix}/with-deleted{trailing_slash}$',
        mapping={
            'get': 'list_with_deleted',
        },
        name='{basename}-with-deleted',
        detail=False,
        initkwargs={'suffix': 'With Deleted'}
    ))

    # bulk restore
    routers.DefaultRouter.routes.insert(2, routers.Route(
        url=r'^{prefix}/restore{trailing_slash}$',
        mapping={
            'patch': 'bulk_restore',
        },
        name='{basename}-restore',
        detail=False,
        initkwargs={'suffix': 'Restore'}
    ))

    # retrieve deleted
    routers.DefaultRouter.routes.insert(5, routers.Route(
        url=r'^{prefix}/deleted/{lookup}{trailing_slash}$',
        name='{basename}-deleted',
        mapping={
            'get': 'retrieve_deleted',
        },
        detail=True,
        initkwargs={'suffix': 'Deleted'}
    ))

    # retrieve with deleted
    routers.DefaultRouter.routes.insert(6, routers.Route(
        url=r'^{prefix}/with-deleted/{lookup}{trailing_slash}$',
        name='{basename}-deleted',
        mapping={
            'get': 'retrieve_with_deleted',
        },
        detail=True,
        initkwargs={'suffix': 'With Deleted'}
    ))

    # restore
    routers.DefaultRouter.routes.insert(8, routers.Route(
        url=r'^{prefix}/{lookup}/restore{trailing_slash}$',
        name='{basename}-restore',
        mapping={
            'patch': 'restore',
        },
        detail=True,
        initkwargs={'suffix': 'Restore'}
    ))



from django.db.models import Manager
from .querysets import SoftDeleteQuerySet


class SoftDeleteManager(Manager):
    def __init__(self, *args, **kwargs):
        self.with_deleted = kwargs.pop('with_deleted', False)
        self.only_deleted = kwargs.pop('only_deleted', False)
        super(SoftDeleteManager, self).__init__(*args, **kwargs)

    def get_queryset(self):
        if self.with_deleted:
            return SoftDeleteQuerySet(self.model)

        if self.only_deleted:
            return SoftDeleteQuerySet(self.model).only_deleted()

        return SoftDeleteQuerySet(self.model).without_deleted()

    def delete(self, hard: bool = False):
        return self.get_queryset().delete(hard=hard)

    def restore(self):
        return self.get_queryset().restore()



import aspectlib

from django.conf import settings
from rest_framework import status, mixins
from rest_framework.generics import GenericAPIView
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.response import Response

from timestamps.drf.utils import is_hard_delete_request


class ListDeletedModelMixin:
    def list_deleted(self, request, *args, **kwargs):
        return ListModelMixin.list(self, request, *args, **kwargs)


class ListWithDeletedModelMixin:
    def list_with_deleted(self, request, *args, **kwargs):
        return ListModelMixin.list(self, request, *args, **kwargs)


class RetrieveDeletedModelMixin:
    def retrieve_deleted(self, request, *args, **kwargs):
        return RetrieveModelMixin.retrieve(self, request, *args, **kwargs)


class RetrieveWithDeletedModelMixin:
    def retrieve_with_deleted(self, request, *args, **kwargs):
        return RetrieveModelMixin.retrieve(self, request, *args, **kwargs)


class RestoreModelMixin:
    def restore(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_restore(instance)

        return Response(self.get_serializer(instance=instance).data)

    def perform_restore(self, instance):
        return instance.restore()


class BulkRestoreModelMixin:
    def bulk_restore(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        count = self.perform_bulk_restore(queryset)

        if getattr(settings, 'BULK_ACTIONS_RETURNS_CONTENT', False):
            return Response(data={'count': count, }, status=status.HTTP_200_OK)

        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_bulk_restore(self, qs):
        return qs.restore()


class DestroyModelMixin(mixins.DestroyModelMixin):
    def perform_destroy(self, instance):
        return instance.delete(hard=is_hard_delete_request(self.request))


class BulkDestroyModelMixin:
    def perform_bulk_destroy(self, qs):
        return qs.delete(hard=is_hard_delete_request(self.request))

    def bulk_destroy(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        count = self.perform_bulk_destroy(queryset)

        if not getattr(settings, 'BULK_ACTIONS_RETURNS_CONTENT', False):
            return Response(status=status.HTTP_204_NO_CONTENT)

        # a delete operation (hard delete) returns a tuple of:
        # - total rows deleted (count)
        # - total rows deleted per table (per_model)
        if isinstance(count, tuple):
            count, count_per_model = count
            return Response(data={'count': count, 'count_per_model': count_per_model, }, status=status.HTTP_200_OK)

        return Response(data={'count': count}, status=status.HTTP_200_OK)


def __remove_clause_deleted_at(queryset):
    from timestamps.querysets import SoftDeleteQuerySet
    from django.db.models.lookups import IsNull

    if not isinstance(queryset, SoftDeleteQuerySet):
        return queryset

    queryset = queryset.all()  # clone
    where = queryset.query.where

    for i, child in enumerate(where.children):
        if isinstance(child, IsNull) and child.lhs.field.name == 'deleted_at':
            where.children.pop(i)
            break

    return queryset


# Using Aspect-Oriented Programming (AOP)
# to change behavior of GenericAPIView.get_queryset(self).
# Doing this way, there is no need to pollute CoreModelViewSet
# and gives to developers the possibility
# to use only a subset of Mixins of soft deleting technology,
# without the need to use all the views mixins, extended in CoreModelViewSet.
@aspectlib.Aspect
def __get_queryset(*args, **kwargs):
    queryset = yield aspectlib.Proceed

    view = args[0]

    mixin = {
        'list_with_deleted': ListWithDeletedModelMixin,
        'retrieve_with_deleted': RetrieveWithDeletedModelMixin,
    }

    if is_hard_delete_request(view.request):
        mixin['destroy'] = DestroyModelMixin
        mixin['bulk_destroy'] = BulkDestroyModelMixin

    mixin = mixin.get(view.action, None)

    if mixin and isinstance(view, mixin):
        queryset = __remove_clause_deleted_at(queryset)
        yield aspectlib.Return(queryset)

    mixin = {
        'list_deleted': ListDeletedModelMixin,
        'retrieve_deleted': RetrieveDeletedModelMixin,
        'restore': RestoreModelMixin,
        'bulk_restore': BulkRestoreModelMixin,
    }.get(view.action, None)

    if mixin and isinstance(view, mixin):
        queryset = __remove_clause_deleted_at(queryset)
        yield aspectlib.Return(queryset.only_deleted())

    yield aspectlib.Return(queryset)


aspectlib.weave(target=GenericAPIView.get_queryset, aspects=__get_queryset)
